@@ -1,6 +1,6 @@
-from client import (Error, NoAgentsError, NoCallbackError, UnknownCommandError,
+from client import (Error, NoCallbackError, UnknownCommandError,
                     Connection, Request, TimeoutError, RequestCancelledError)
-
+from shard import NoAgentsError
 
 __all__ = ['Error', 'NoAgentsError', 'NoCallbackError', 'UnknownCommandError',
            'Connection', 'Request', 'TimeoutError', 'RequestCancelledError']
