@@ -24,19 +24,50 @@ To use the Notch Client, first create a Client instance::
 
   conn = notch.client.Connection('localhost:8800')
 
-If you have multiple agents to connect to, supply a list to :class:`Connection`::
+If you have multiple agents, supply a list to :class:`Connection`::
 
   conn = notch.client.Connection(['localhost:8800', 'localhost:8801'])
 
-First, create a Client instance then call Notch API methods and
-provide callbacks to receive the results after the I/O operations
-complete.
+Or alternatively, if you have many agents, and would like to lay them out
+according to shards based on the device name in a request (e.g., for
+geographical distribution), you can create a file ~/.notch_shards, or
+/etc/notch_shards (the one in $HOME takes preference).  This file is YAML,
+and lists shard definitions, one per line. The left side is a key, a regular
+expression matched against a request's device name, followed by a string which
+is the list of agents host:port pairs to connect to for that shard.  e.g.,
 
-If you have listed your agents in the ``NOTCH_AGENTS`` environment variable,
-you can create a client as simply as::
+.+\.lax01: a1-11.lax01:8000,a1-11.lax01:80001,a1-22.lax01:8000,a1-22.lax01:8001
+.+\.syd02: nms1.syd02:8000,nms1.syd02:80001,nms2.syd02:8000,nms2.syd02:8001
+__default__: nms1.syd02:8000,a1-11.lax01:8001
+
+A device with the name "pr11.lax01" would use the agents on the first line,
+one named "cr4.syd02" will use the second line's agents, and devices not
+matching those two would use the two agents on the final line.
+
+You should define a __default__ key, which lists the agents to use for
+requests that have no specific device_name (e.g., device lookups), or
+those which don't match any other line.  It is best to use __default__
+rather than a regular expression like ".*".
+
+All regular expressions are prefixed with ^, hence using .+\.XXX in the
+examples above.
+
+If you don't define a __default__ key, any random shard you've defined
+will be used as the default shard.  It's better to pick a geographically
+distributed shard.
+
+If you used the shard map, or if you have listed your agents in the
+``NOTCH_AGENTS`` environment variable, you can create a client as simply as::
 
   conn = notch.client.Connection()
 
+A value in ``NOTCH_AGENTS`` or arguments supplied in the Connection()
+initialiser will cause the shard map to NOT be loaded.
+
+Next, with your Client instance, call Notch API methods and provide callbacks
+to receive the results after the I/O operations complete, or work in
+synchronous mode.
+
 Synchronous mode
 ----------------
 
@@ -116,24 +147,16 @@ For example::
   conn = notch.client.Connection(load_balancing_policy='RoundRobin')
 
 """
-
-import base64
-import copy
 import eventlet
-
-import logging
-import os
-import traceback
 from eventlet.green import socket
 
-try:
-    import jsonrpclib
-    transport_class = jsonrpclib.jsonrpc.Transport
-except Exception, e:
-    print e.__class__.__name__, str(e)
-    import jsonrpclib
-    transport_class = jsonrpclib.Transport
-    
+
+import base64
+import logging
+import os
+
+import jsonrpclib
+
 # Disable implicit class conversion, as it only
 # really 'works' for built-in types and classes in the
 # standard library (not our objects) without prior
@@ -144,17 +167,14 @@ jsonrpclib.config.use_jsonclass = False
 # by jsonrpclib.
 eventlet.monkey_patch(all=False, socket=True)
 
-
-import lb_transport
-
-import errors
-
+import notch.client.lb_transport
+import notch.client.errors
+import notch.client.shard
 
 # Use at most this number of green threads. Default also influenced by
 # setting NOTCH_CONCURRENCY environment variable.
 DEFAULT_NOTCH_CONCURRENCY = 50
 
-
 # Error classes.
 
 class Error(Exception):
@@ -169,10 +189,6 @@ class TimeoutError(Error):
     """The client request timed out."""
 
 
-class NoAgentsError(Error):
-    """No agent addresses were passed to the client."""
-
-
 class NoCallbackError(Error):
     """The client got an response for a request but has no callback to run."""
 
@@ -224,9 +240,9 @@ class Request(object):
         self.time_completed = None
         self.time_elapsed_s = None
 
-    valid = property(lambda x: (x.notch_method and x.arguments))
-    completed = property(lambda x: x._completed())
-    is_async = property(lambda x: bool(x.callback))
+    valid = property(lambda self: (self.notch_method and self.arguments))
+    completed = property(lambda self: self._completed())
+    is_async = property(lambda self: bool(self.callback))
 
     def __copy__(self):
         """Produce a deep copy of the object."""
@@ -254,7 +270,7 @@ class Request(object):
                      or self.error is not None)
                     and not self._timeout.pending)
         else:
-            return (self.result is not None or self.error is not None)
+            return self.result is not None or self.error is not None
 
     def start(self):
         if self.timeout_s is not None:
@@ -327,12 +343,8 @@ class Connection(object):
           use_ssl: A boolean, if True, use HTTPS instead of HTTP for transport.
 
         Raises:
-          NoAgentsError: If no agents were supplied.
+          shard.NoAgentsError: If no agents were supplied.
         """
-        if isinstance(agents, str):
-            self.agents = [agents]
-        else:
-            self.agents = agents or os.getenv('NOTCH_AGENTS')
         try:
             self.max_concurrency = int(max_concurrency or
                                        os.getenv('NOTCH_CONCURRENCY') or
@@ -350,10 +362,11 @@ class Connection(object):
         self._lb_policy = None
         if load_balancing_policy is not None:
             self.load_balancing_policy = load_balancing_policy
+
+        self._shard_manager = notch.client.shard.AgentShardManager(
+            self._lb_policy)
+        self._shard_manager.setup(agents)
         self._pool = eventlet.greenpool.GreenPool(self.max_concurrency)
-        self._notch = None
-        self._transport = None
-        self._setup_agents()
 
     @property
     def counters(self):
@@ -364,8 +377,8 @@ class Connection(object):
 
     def _set_load_balancing_policy(self, lbp):
         """Selects the new load balancing policy and activates it."""
-        if hasattr(lb_transport, lbp):
-            self._lb_policy = getattr(lb_transport, lbp)
+        if hasattr(notch.client.lb_transport, lbp):
+            self._lb_policy = getattr(notch.client.lb_transport, lbp)
         else:
             raise NoSuchLoadBalancingPolicyError(
                 'There is no load balancing policy named %r' % lbp)
@@ -373,7 +386,7 @@ class Connection(object):
 
     load_balancing_policy = property(_load_balancing_policy,
                                      _set_load_balancing_policy)
-                                     
+
     def _request_callback(self, gt, *args, **kwargs):
         """Asynchronously receives responses and runs the user callback.
 
@@ -409,8 +422,10 @@ class Connection(object):
                     'Asynchronous mode used without defined callback.')
 
     def _send_notch_api_request(self, method, request, kwarg_list):
+        device_name = request.arguments.get('device_name')
+        notch, _ = self._shard_manager.shard_for_device(device_name)
         try:
-            func = getattr(self._notch, method)
+            func = getattr(notch, method)
             args = dict((kwarg, request.arguments.get(kwarg, None)) for
                         kwarg in kwarg_list)
             request.result = func(**args)
@@ -432,16 +447,20 @@ class Connection(object):
                 sections = str(e).split(':')
                 try:
                     err_no = int(sections[0].split()[1])
-                    err_name = (errors.reverse_error_dictionary
-                                .get(err_no,'Unknown'))
-                    e.args = (err_name, str(e))
-                except (KeyError, ValueError), _:
-                    err_name = 'Unknown'
+                    err_name = (
+                        notch.client.errors.reverse_error_dictionary.get(
+                            err_no))
+                    err_class = getattr(notch.client.errors, err_name)
+                    if err_class is not None:
+                        new_e = err_class(sections[1])
+                    else:
+                        new_e = jsonrpclib.ProtocolError(err_no, sections[1])
+                    print str(new_e)
+                except (KeyError, ValueError):
+                    new_e = notch.client.errors.ApiError(str(e))
             else:
-                err_name = e.__class__.__name__
-                err_msg = str(e)
-                e.args = (err_name, str(e))
-            request.error = e
+                new_e = e
+            request.error = new_e
         return request
 
     def _notch_api_command(self, request):
@@ -466,28 +485,13 @@ class Connection(object):
         return self._send_notch_api_request(
             'get_config', request, ('device_name', 'source', 'mode'))
 
-    def _setup_agents(self):
-        """Sets up the Notch JSON RPC agent connection.
-
-        Raises:
-          NoAgentsError: If no agents were supplied.
-        """
-        if not self.agents:
-            raise NoAgentsError('No Notch agents supplied to Client.')
-        else:
-            self._transport = lb_transport.LoadBalancingTransport(
-                hosts=self.agents, transport=transport_class,
-                policy=self._lb_policy)
-            self._notch = jsonrpclib.Server(
-                self._protocol + str(self.path), transport=self._transport)
-
     def _exec_requests(self, requests):
-        """Executes a iterable of requests, returning responses or greenthreads.
+        """Executes a iterable of requests, returns responses or greenthreads.
 
         If requests is an iterable of mixed synchronous and
         asynchronous Request objects, this method's return
         value is undefined.
-        
+
         Args:
           requests: An iterable of Request objects.
 
@@ -554,9 +558,12 @@ class Connection(object):
           can be wait()ed on.
           Inspect the result or error attributes of the request object.
         """
-        if callback: request.callback = callback
-        if args: request.callback_args = args
-        if kwargs: request.callback_kwargs = kwargs
+        if callback:
+            request.callback = callback
+        if args:
+            request.callback_args = args
+        if kwargs:
+            request.callback_kwargs = kwargs
         result = self._exec_requests([request])
         # We always just want the first (result or gt) from the list.
         return result[0]
@@ -578,9 +585,12 @@ class Connection(object):
           kwargs: Dict of keyword arguments for the user callback. Optional.
         """
         for request in requests:
-            if callback: request.callback = callback
-            if args: request.callback_args = args
-            if kwargs: request.callback_kwargs = kwargs
+            if callback:
+                request.callback = callback
+            if args:
+                request.callback_args = args
+            if kwargs:
+                request.callback_kwargs = kwargs
         return self._exec_requests(requests)
 
     def kill_all(self):
@@ -596,8 +606,8 @@ class Connection(object):
         if hasattr(self, '_pool') and self._pool is not None:
             self._pool.waitall()
 
-    num_requests_running = property(lambda x: x._pool.running())
-    num_requests_waiting = property(lambda x: x._pool.waiting())
+    num_requests_running = property(lambda self: self._pool.running())
+    num_requests_waiting = property(lambda self: self._pool.waiting())
 
     # xmlrpclib style Proxy methods. If one initialises
     # notch.client.Client instead of an xmlrpclib server proxy, these
@@ -693,7 +703,7 @@ class Connection(object):
                           callback=callback, callback_args=callback_args,
                           callback_kwargs=callback_kwargs)
         r = self.exec_request(request)
-        
+
         if request.callback:
             return
         else:
@@ -710,7 +720,7 @@ class Connection(object):
                           callback=callback, callback_args=callback_args,
                           callback_kwargs=callback_kwargs)
         r = self.exec_request(request)
-        
+
         if request.callback:
             return
         else:
